/*----------------------------------------------------------------------
  File    : dialog.c
  Contents: simplified dialog definition for Athena widgets
            (action functions, alert/about dialog, property sheets)
  History : 15.05.1998 file created as psheet.c
            03.06.1998 minor improvements
            06.08.1998 functions ps_focus and ps_next added
            09.01.2000 callback function stored as client data
            12.01.2000 action functions and alert/about dialog added
            15.01.2000 minimal width for toggles (checkboxes) added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include "dialog.h"
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PS_LINE     -1          /* line label */
#define BLKSIZE     32          /* block size for item vector */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "excl.xbm"             /* exclamation mark for alert boxes */
#include "logo.xbm"             /* logo bitmap (Otto von Guericke) */

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtActionsRec actions[] = {
  { "db_focus",  db_focus },    /* set focus on input field */
  { "db_next",   db_next  },    /* switch to next input field */
  { "db_close",  db_close } };  /* close dialog box */

/*----------------------------------------------------------------------
  Dialog Box Action Functions
----------------------------------------------------------------------*/

void db_focus (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- set focus on input field */
  XtSetKeyboardFocus(XtParent(XtParent(w)), w);
}  /* db_focus() */

/*--------------------------------------------------------------------*/

void db_next (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- switch to next input field */
  Widget     form;              /* form widget of property sheet */
  WidgetList list;              /* list of child widgets */
  Cardinal   n, i;              /* number of child widgets, index */
  Boolean    dc;                /* whether to display caret */

  form = XtParent(w);           /* get form widget */
  XtVaGetValues(form, XtNnumChildren, &n, XtNchildren, &list, NULL);
  for (i = 0; i < n; i++)       /* get children of shell widget */
    if (list[i] == w) break;    /* and look for current widget */
  if (i >= n) return;           /* if not found, abort function */
  while (1) {                   /* widget search loop */
    if (++i >= n) i = 0;        /* go to next child of shell widget, */
    if (list[i] == w) break;    /* if round completed, abort */
    if (XtClass(list[i]) != asciiTextWidgetClass)
      continue;                 /* skip non-text gadgets */
    XtVaGetValues(list[i], XtNdisplayCaret, &dc, NULL);
    if (dc) break;              /* if to display a caret (editable), */
  }                             /* abort the loop */
  XtSetKeyboardFocus(XtParent(form), list[i]);
}  /* db_next() */              /* set new input focus */

/*--------------------------------------------------------------------*/

void db_close (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- close dialog box */
  Widget b;                     /* buffer for button widget */

  while (!XtIsTransientShell(w))
    w = XtParent(w);            /* find the shell widget */
  if ((*c == 0) || (strcmp(*s, "cancel") == 0)) {
    b = XtNameToWidget(w, "*cancel");
    if (b) { XtCallCallbacks(b, XtNcallback, NULL); return; }
  }                             /* call button callback function */
  b = XtNameToWidget(w, "*ok"); /* of a cancel or an ok button */
  if (b) { XtCallCallbacks(b, XtNcallback, NULL); return; }
  XtPopdown(w);                 /* close the dialog box */
}  /* db_close() */

/*----------------------------------------------------------------------
  Alert Box Function
----------------------------------------------------------------------*/

int db_alert (Widget parent, XtCallbackProc callback, const char *msg)
{                               /* --- show an alert dialog box */
  static Widget alert = NULL;   /* shell widget */
  static int    init  = 0;      /* initialization flag */
  Widget    shell, form;        /* shell and form for alert box */
  Widget    text, ok;           /* widgets of message and button */
  Pixmap    excl;               /* bitmap for exclamation mark */
  Arg       args[4];            /* widget arguments */
  int       n;                  /* argument counter */
  Position  x,  y;              /* position of dialog box */
  Dimension wd, ht;             /* width and height of created dialog */
  Atom      wm_delwin;          /* delete window atom of window mgr. */

  if (!alert) {                 /* if widget is not created */
    shell = XtCreatePopupShell("alertDlg", transientShellWidgetClass,
              parent, NULL, 0); /* create a dialog box shell */
    if (!shell) { XtDestroyWidget(shell); return -1; }
    XtVaSetValues(shell, XtNallowShellResize, True, NULL);
    XtSetArg(args[0], XtNborderWidth, 0);
    form = XtCreateManagedWidget("form", formWidgetClass,
             shell, args, 1);   /* create form for box layout */
    if (!form)  { XtDestroyWidget(shell); return -1; }
    excl  = XCreateBitmapFromData(XtDisplay(parent), XtWindow(parent),
              (char*)excl_bits, excl_width, excl_height);
    if (!excl) { XtDestroyWidget(shell); return -1; }
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNborderWidth, 0);    n++;
    XtSetArg(args[n], XtNresizable,   True); n++;
    XtSetArg(args[n], XtNleftBitmap,  excl); n++;
    text = XtCreateManagedWidget("msg", labelWidgetClass,
              form, args, n);   /* create text label */
    if (!text) { XtDestroyWidget(shell); return -1; }
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromVert,  text);   n++;
    XtSetArg(args[n], XtNresizable, True);   n++;
    ok   = XtCreateManagedWidget("ok", commandWidgetClass,
             form, args, n);    /* create 'ok' button */
    if (!ok) { XtDestroyWidget(shell); return -1; }
    XtInstallAllAccelerators(form, form);
    alert = shell;              /* install accelerators and */
  }                             /* note created dialog box widget */
  if (!msg) return 0;           /* if no message, abort function */
  text = XtNameToWidget(alert, "*msg");
  ok   = XtNameToWidget(alert, "*ok");
  XtRemoveAllCallbacks(ok, XtNcallback);
  XtAddCallback(ok,     XtNcallback, callback, NULL);
  XtVaSetValues(text,   XtNlabel, msg, NULL);
  XtVaGetValues(text,   XtNwidth, &wd, XtNheight, &ht, NULL);
  if (ht < excl_height)         /* adapt width and height of widgets */
    XtVaSetValues(text, XtNheight, excl_height, NULL);
  XtVaSetValues(ok,     XtNwidth, wd,  NULL);
  XtVaGetValues(parent, XtNx, &x, XtNy, &y, NULL);
  XtVaSetValues(alert,  XtNx,  x, XtNy,  y, NULL);
  XtPopup(alert, XtGrabExclusive);  /* place and show alert box */
  XtVaGetValues(alert,  XtNwidth, &wd, XtNheight, &ht, NULL);
  XtVaSetValues(alert,  XtNminWidth, wd, XtNminHeight, ht,
                        XtNmaxWidth, wd, XtNmaxHeight, ht, NULL);
  if (!init) {                  /* if shell was created */
    XtOverrideTranslations(alert,
      XtParseTranslationTable("<Message>WM_PROTOCOLS: db_close()"));
    wm_delwin = XInternAtom(XtDisplay(alert),
      "WM_DELETE_WINDOW", False);
    XSetWMProtocols(XtDisplay(alert), XtWindow(alert), &wm_delwin, 1);
    init = 1;                   /* override WM delete and */
  }                             /* set the initialization flag */
  return 0;                     /* return `ok' */ 
}  /* db_alert() */

/*----------------------------------------------------------------------
  About Box Function
----------------------------------------------------------------------*/

int db_about (Widget parent, XtCallbackProc callback)
{                               /* --- show an about dialog box */
  static Widget about = NULL;   /* shell widget of about box */
  Widget    shell = NULL, form; /* widgets for shell and form */
  Widget    label, logo, ok;    /* widgets for labels and button */
  Pixmap    bmap;               /* logo bitmap */
  Arg       args[4];            /* widget argument list */
  int       n;                  /* argument counter */
  Position  x,  y;              /* position of dialog box */
  Dimension wd, ht;             /* width and height of created dialog */
  Atom      wm_delwin;          /* delete window atom of window mgr. */

  if (!about) {                 /* if widget is not created */
    shell = XtCreatePopupShell("aboutDlg", transientShellWidgetClass,
              parent, NULL, 0); /* create a dialog box shell */
    if (!shell) { XtDestroyWidget(shell); return -1; }
    XtSetArg(args[0], XtNborderWidth, 0);
    form  = XtCreateManagedWidget("form", formWidgetClass,
              shell, args, 1);   /* create form for box layout */
    if (!form)  { XtDestroyWidget(shell); return -1; }
    XtSetArg(args[0], XtNborderWidth, 0);
    label = XtCreateManagedWidget("info", labelWidgetClass,
              form, args, 1);   /* create text label */
    if (!label) { XtDestroyWidget(shell); return -1; }
    bmap  = XCreateBitmapFromData(XtDisplay(parent), XtWindow(parent),
              (char*)logo_bits, logo_width, logo_height);
    if (bmap == None) {         /* create logo bitmap */
      XtDestroyWidget(shell); return -1; }
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNborderWidth,   0);     n++;
    XtSetArg(args[n], XtNinternalWidth, 0);     n++;
    XtSetArg(args[n], XtNfromHoriz,     label); n++;
    XtSetArg(args[n], XtNbitmap,        bmap);  n++;
    logo = XtCreateManagedWidget("logo", labelWidgetClass,
             form, args, n);    /* create logo label */
    if (!logo) { XtDestroyWidget(shell);
      XFreePixmap(XtDisplay(parent), bmap); return -1; }
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromHoriz, label);      n++;
    XtSetArg(args[n], XtNfromVert,  logo);       n++;
    XtSetArg(args[n], XtNwidth,     logo_width); n++;
    ok = XtCreateManagedWidget("ok", commandWidgetClass,
           form, args, n);      /* create 'ok' button */
    if (!ok) { XtDestroyWidget(shell);
      XFreePixmap(XtDisplay(parent), bmap); return -1; }
    XtInstallAllAccelerators(form, form);
    about = shell;              /* install accelerators and */
  }                             /* note created dialog box widget */
  ok = XtNameToWidget(about, "*ok");
  XtRemoveAllCallbacks(ok, XtNcallback);
  XtAddCallback(ok, XtNcallback, callback, NULL);
  XtVaGetValues(parent, XtNx, &x, XtNy, &y, NULL);
  XtVaSetValues(about,  XtNx,  x, XtNy,  y, NULL);
  XtPopup(about, XtGrabExclusive);
  if (shell) {                  /* show and prepare dialog box */
    XtVaGetValues(shell, XtNwidth,   &wd, XtNheight,   &ht, NULL);
    XtVaSetValues(shell, XtNminWidth, wd, XtNminHeight, ht,
                         XtNmaxWidth, wd, XtNmaxHeight, ht, NULL);
    XtOverrideTranslations(shell,
      XtParseTranslationTable("<Message>WM_PROTOCOLS: db_close()"));
    wm_delwin = XInternAtom(XtDisplay(shell),
      "WM_DELETE_WINDOW", False);
    XSetWMProtocols(XtDisplay(shell), XtWindow(shell), &wm_delwin, 1);
  }                             /* inhibit resize, override WM delete */
  return 0;                     /* return 'ok' */
}  /* db_about() */

/*----------------------------------------------------------------------
  Callback Functions
----------------------------------------------------------------------*/

static void cb_ok (Widget w, XtPointer client, XtPointer call)
{                               /* --- callback for ok button */
  XtPopdown(XtParent(XtParent(w)));  /* close dialog box */
  ((XtCallbackProc)client)(w, (void*)1, NULL);
}  /* cb_ok() */

/*--------------------------------------------------------------------*/

static void cb_cancel (Widget w, XtPointer client, XtPointer call)
{                               /* --- callback for cancel button */
  XtPopdown(XtParent(XtParent(w)));  /* close dialog box */
  ((XtCallbackProc)client)(w, (void*)0, NULL);
}  /* cb_cancel() */

/*--------------------------------------------------------------------*/

static void cb_menu (Widget w, XtPointer client, XtPointer call)
{                               /* --- callback for menu selection */
  String text;                  /* text of menu entry */
  XtVaGetValues(w, XtNlabel, &text, NULL);
  XtVaSetValues(XtParent(XtParent(w)), XtNlabel, text, NULL);
}  /* cb_menu() */              /* set text of selected menu item */

/*----------------------------------------------------------------------
  Property Sheet Functions
----------------------------------------------------------------------*/

PSHEET* ps_create (const char *name, Widget parent, int cabutton)
{                               /* --- create a property sheet */
  PSHEET *psh;                  /* created property sheet */
  Arg    args[1];               /* widget argument list */

  psh = (PSHEET*)malloc(sizeof(PSHEET));
  if (!psh) return NULL;        /* allocate memory and */
  psh->parent   = parent;       /* initialize fields */
  psh->itemvsz  = psh->itemcnt  = 0;
  psh->items    = NULL;
  psh->currline = psh->prevline = psh->previtem = NULL;
  psh->shell    = XtCreatePopupShell(name, transientShellWidgetClass,
    parent, NULL, 0);           /* create a dialog box shell */
  if (!psh->shell) { ps_delete(psh); return NULL; }
  XtSetArg(args[0], XtNborderWidth, 0);
  psh->form = XtCreateManagedWidget("form", formWidgetClass,
    psh->shell, args, 1);       /* create a dialog box form */
  if (!psh->form) { ps_delete(psh); return NULL; }
  psh->ok = XtCreateManagedWidget("ok", commandWidgetClass,
    psh->form, args, 0);        /* create 'ok' button */
  if (!psh->ok) { ps_delete(psh); return NULL; }
  if (!cabutton) {              /* if no cancel button requested, */
    psh->ca = NULL; return psh; }           /* abort the function */
  XtSetArg(args[0], XtNfromHoriz, psh->ok);
  psh->ca = XtCreateManagedWidget("cancel", commandWidgetClass,
    psh->form, args, 1);        /* create 'cancel' button */
  if (!psh->ca) { ps_delete(psh); return NULL; }
  return psh;                   /* return created property sheet */
}  /* ps_create() */

/*--------------------------------------------------------------------*/

void ps_delete (PSHEET *psh)
{                               /* --- delete a property sheet */
  if (psh->shell) XtDestroyWidget(psh->shell);
  if (psh->items) free(psh->items);  /* destroy widgets and */
  free(psh);                         /* delete property sheet */
}  /* ps_delete() */

/*--------------------------------------------------------------------*/

int ps_addline (PSHEET *psh, const char *name)
{                               /* --- add a line to a prop. sheet */
  int    vsz;                   /* size of item vector */
  PSITEM *tmp;                  /* temporary buffer */
  Arg    args[4];               /* widget argument list */
  int    n;                     /* argument counter */
  Widget label = NULL;          /* widget for line label */

  vsz = psh->itemvsz;           /* get the label vector size */
  if (psh->itemcnt >= vsz) {    /* if the label vector is full, */
    vsz += BLKSIZE;             /* enlarge the item vector */
    tmp = (PSITEM*)realloc(psh->items, vsz *sizeof(PSITEM));
    if (!tmp) return -1;        /* allocate (new) item vector */
    psh->items = tmp; psh->itemvsz = vsz;
  }                             /* set (new) item vector and its size */
  if (name) {                   /* if label name is given */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNborderWidth, 0);             n++;
    XtSetArg(args[n], XtNjustify,     XtJustifyLeft); n++;
    XtSetArg(args[n], XtNfromVert,    psh->currline); n++;
    label = XtCreateManagedWidget(name, labelWidgetClass,
      psh->form, args, n);      /* create a label widget */
    if (!label) return -1;
  }                             /* note label widget */
  tmp = psh->items +psh->itemcnt++;
  tmp->widget = label; tmp->type = PS_LINE;
  psh->prevline = psh->currline;
  psh->currline = psh->previtem = label;
  return 0;                     /* return 'ok' */
}  /* ps_addline() */

/*--------------------------------------------------------------------*/

int ps_additem (PSHEET *psh, const char *name, int type, int width)
{                               /* --- add an item to a prop. sheet */
  XtTranslations tl;            /* translations for text widgets */
  int         vsz;              /* size of item vector */
  PSITEM      *tmp;             /* temporary buffer */
  Arg         args[6];          /* widget argument list */
  int         n;                /* argument counter */
  WidgetClass class;            /* class of created item */
  Widget      item;             /* widget for created item */
  Widget      pane;             /* widget for menu pane */
  String      s;                /* label of toggle gadget */

  vsz = psh->itemvsz;           /* get the label vector size */
  if (psh->itemcnt >= vsz) {    /* if the label vector is full, */
    vsz += BLKSIZE;             /* enlarge the item vector */
    tmp = (PSITEM*)realloc(psh->items, vsz *sizeof(PSITEM));
    if (!tmp) return -1;        /* allocate (new) item vector */
    psh->items = tmp; psh->itemvsz = vsz;
  }                             /* set (new) item vector and its size */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromHoriz, psh->previtem); n++;
  XtSetArg(args[n], XtNfromVert,  psh->prevline); n++;
  switch (type) {               /* evaluate item type */
    case PS_LABEL  :            /* label */
      XtSetArg(args[n], XtNborderWidth,  0);             n++;
      XtSetArg(args[n], XtNjustify,      XtJustifyLeft); n++;
      class = labelWidgetClass;      break;
    case PS_COMMAND:            /* command button */
      XtSetArg(args[n], XtNborderWidth,  1);             n++;
      class = commandWidgetClass;    break;
    case PS_TEXT   :            /* non-editable text field */
      XtSetArg(args[n], XtNborderWidth,  1);             n++;
      XtSetArg(args[n], XtNeditType,     XawtextRead);   n++;
      XtSetArg(args[n], XtNdisplayCaret, False);         n++;
      XtSetArg(args[n], XtNcursor,       XC_X_cursor);   n++;
      class = asciiTextWidgetClass;  break;
    case PS_EDIT   :            /* editable text field */
      XtSetArg(args[n], XtNborderWidth,  1);             n++;
      XtSetArg(args[n], XtNeditType,     XawtextEdit);   n++;
      class = asciiTextWidgetClass;  break;
    case PS_TOGGLE :            /* toggle button (checkbox) */
      XtSetArg(args[n], XtNborderWidth,  1);             n++;
      class = toggleWidgetClass;     break;
    case PS_RADIO  :            /* radio button */
      XtSetArg(args[n], XtNborderWidth,  1);             n++;
      class = toggleWidgetClass;     break;
    case PS_MENU   :            /* menu button */
      XtSetArg(args[n], XtNborderWidth,  1);             n++;
      XtSetArg(args[n], XtNmenuName,     name);          n++;
      class = menuButtonWidgetClass; break;
    default        :            /* an unknown item type */
      return -1;                /* abort function */
  }
  item = XtCreateManagedWidget(name, class, psh->form, args, n);
  if (!item) return -1;         /* create a widget */
  if (width > 0)                /* if width given, set it */
    XtVaSetValues(item, XtNwidth, width, NULL);
  if      (type == PS_MENU) {   /* if the object is a menu */
    pane = XtCreatePopupShell(name, simpleMenuWidgetClass,
             item, NULL, 0);    /* create a menu pane */
    if (!pane) { XtDestroyWidget(item); return -1; } }
  else if (type == PS_TOGGLE) { /* if the object is a checkbox */
    if (width <= 0) {           /* if no width given */
      XtVaGetValues(item, XtNlabel, &s, NULL);
      if (*s == '\0') XtVaSetValues(item, XtNwidth, 10, NULL);
    } }                         /* set minimal width of toggle */
  else if (type == PS_EDIT) {   /* if the object is a text gadget */
    tl = XtParseTranslationTable(
           "<Key>Home  : beginning-of-line()\n"
           "<Key>End   : end-of-line()\n");
    XtAugmentTranslations(item, tl);
    tl = XtParseTranslationTable(
           "<Key>Tab   : db_next()\n"
           "<Key>Return: db_close(ok)\n"
           "!Alt<Key>o : db_close(ok)\n"
           "!Alt<Key>c : db_close(cancel)\n"
           "<Btn1Down> : db_focus()\n");
    XtOverrideTranslations(item, tl);
  }                             /* augment/override translations */
  tmp = psh->items +psh->itemcnt++;
  tmp->widget = psh->previtem = item;
  tmp->type   = type;           /* note item and its type */
  if (!psh->currline) psh->currline = item;
  return 0;                     /* return `ok' */
}  /* ps_additem() */

/*--------------------------------------------------------------------*/

int ps_addcb (PSHEET *psh, const char *name,
              XtCallbackProc callback, XtPointer data)
{                               /* --- set callback procedure */
  XtAddCallback(XtNameToWidget(psh->form, name),
    XtNcallback, callback, data);
  return 0;                     /* return `ok' */
}  /* ps_addcb() */

/*--------------------------------------------------------------------*/

int ps_menuadd (PSHEET *psh, const char *name, const char *text)
{                               /* --- add an item to a menu */
  Widget pane;                  /* widget for menu pane */
  Widget item;                  /* widget for menu item */

  pane = XtNameToWidget(psh->form, name);
  pane = XtNameToWidget(pane, name);
  if (!text) {                  /* if no text is given */
    item = XtCreateManagedWidget("_line_",
             smeLineObjectClass, pane, NULL, 0);
    if (!item) return -1; }     /* create a separating line */
  else {                        /* if an item name is given */
    item = XtCreateManagedWidget(text,
             smeBSBObjectClass, pane, NULL, 0);
    if (!item) return -1;       /* create a menu item */
    XtAddCallback(item, XtNcallback, cb_menu, NULL);
  }                             /* add callback function */
  return 0;                     /* return `ok' */
}  /* ps_menuadd() */

/*--------------------------------------------------------------------*/

int ps_setval (PSHEET *psh, const char *name, const char *text)
{                               /* --- set a prop. sheet item value */
  int    i, k;                  /* loop variable, buffer */
  Widget item;                  /* widget of property sheet item */
  Widget *mnitem;               /* to traverse menu items */
  PSITEM *psi;                  /* to traverse item vector */
  String menu;                  /* name of menu */
  String label;                 /* label */

  item = XtNameToWidget(psh->form, name);
  psi  = psh->items;            /* traverse property sheet items */
  for (i = psh->itemcnt; --i >= 0; psi++)
    if (item == psi->widget) break;
  if (i < 0) return -1;         /* find item in item vector */
  switch (psi->type) {          /* according to item type */
    case PS_TOGGLE:             /* toggle button (checkbox) */
      XtVaSetValues(item, XtNstate, (text) ? True : False, NULL);
      break;                    /* set new state of toggle */
    case PS_TEXT:               /* non-editable text field */
    case PS_EDIT:               /* editable text field */
      XtVaSetValues(item, XtNstring, text, NULL);
      break;                    /* set new string */
    case PS_MENU:               /* menu */
      XtVaGetValues(item, XtNmenuName, &menu, NULL);
      XtVaGetValues(XtNameToWidget(psi->widget, menu),
        XtNnumChildren, &k, XtNchildren, &mnitem, NULL);
      i = (int)text;            /* get index of selection */
      if (i >= k) return -1;    /* and check it */
      XtVaGetValues(mnitem[i], XtNlabel, &label, NULL);
      XtVaSetValues(item,      XtNlabel, label,  NULL);
      break;                    /* set new selection */
    default:                    /* otherwise */
      XtVaSetValues(item, XtNlabel, text, NULL);
      break;                    /* set new label */
  }
  return 0;                     /* return `ok' */
}  /* ps_setval() */

/*--------------------------------------------------------------------*/

const char* ps_getval (PSHEET *psh, const char *name)
{                               /* --- get a prop. sheet item value */
  int     i, k;                 /* loop variables */
  Widget  item;                 /* widget of property sheet item */
  Widget  *mnitem;              /* to traverse menu items */
  PSITEM  *psi;                 /* to traverse item vector */
  String  menu;                 /* name of menu */
  String  label;                /* label */
  String  text;                 /* text input */
  Boolean state;                /* state of toggle */

  item = XtNameToWidget(psh->form, name);
  psi  = psh->items;            /* traverse property sheet items */
  for (i = psh->itemcnt; --i >= 0; psi++)
    if (item == psi->widget) break;
  if (i < 0) return NULL;       /* find item in item vector */
  switch (psi->type) {          /* according to item type */
    case PS_TOGGLE:             /* toggle button (checkbox) */
      XtVaGetValues(item, XtNstate, &state, NULL);
      return (const char*)((state) ? 1 : 0);
                                /* get and return state of toggle */
    case PS_TEXT:               /* non-editable text field */
    case PS_EDIT:               /* editable text field */
      XtVaGetValues(item, XtNstring, &text, NULL);
      return text;              /* get and return input text */
    case PS_MENU:               /* menu */
      XtVaGetValues(item, XtNmenuName, &menu, NULL);
      XtVaGetValues(XtNameToWidget(psi->widget, menu),
        XtNnumChildren, &k, XtNchildren, &mnitem, NULL);
      XtVaGetValues(item, XtNlabel, &label, NULL);
      for (i = 0; i < k; i++) { /* traverse menu items */
        XtVaGetValues(mnitem[i], XtNlabel, &text, NULL);
        if (strcmp(label, text) == 0) return (const char*)i;
      } return (const char*)-1; /* get and return index of selection */
    default:                    /* otherwise */
      XtVaGetValues(item, XtNlabel, &label, NULL);
      return label;             /* get and return label of item */
  }
}  /* ps_getval() */

/*--------------------------------------------------------------------*/

int ps_handle (PSHEET *psh, XtCallbackProc callback)
{                               /* --- handle property sheet dialog */
  Atom      wm_delete;          /* delete window atom */
  Dimension wd, ht;             /* width and height of created dialog */
  Dimension dd;                 /* default distance of controls */
  Dimension max;                /* maximal width of labels/controls */
  Dimension total;              /* total width of dialog box */
  Dimension tmp;                /* temporary buffer */
  Position  x, y;               /* position of dialog box */
  Arg       args[2];            /* widget argument list */
  int       n, i;               /* argument counter, loop variable */
  PSITEM    *psi;               /* to traverse items */
  Widget    *item;              /* to traverse menu items */
  String    name;               /* name of menu */

  /* --- determine and set label width --- */
  psi = psh->items; max = 0;    /* traverse property sheet items */
  for (n = psh->itemcnt; --n >= 0; psi++) {
    if ((psi->type != PS_LINE)  /* skip non-labels */
    ||  !psi->widget) continue; /* and empty labels */
    XtVaGetValues(psi->widget, XtNwidth, &wd, NULL);
    if (wd > max) max = wd;     /* determine the maximum */
  }                             /* of the label widths */
  psi = psh->items;             /* traverse property sheet items */
  for (n = psh->itemcnt; --n >= 0; psi++) {
    if ((psi->type != PS_LINE)  /* skip non-labels */
    ||  !psi->widget) continue; /* and empty labels */
    XtVaSetValues(psi->widget, XtNwidth, max, NULL);
  }                             /* set maximal label width */
  total = max;                  /* and note it */

  /* --- determine menu widths --- */
  psi = psh->items;             /* traverse property sheet items */
  for (n = psh->itemcnt; --n >= 0; psi++) {
    if (psi->type != PS_MENU) continue;   /* skip non-menus */
    XtVaGetValues(psi->widget, XtNwidth, &max,
      XtNmenuName, &name, NULL);
    XtVaGetValues(XtNameToWidget(psi->widget, name),
      XtNnumChildren, &i, XtNchildren, &item, NULL);
    for ( ; --i >= 0; item++) {
      XtVaGetValues(*item, XtNwidth, &wd, NULL);
      if (wd > max) max = wd;   /* determine maximal width */
    }                           /* of menu entries */
    XtVaSetValues(psi->widget, XtNwidth, max, NULL);
  }                             /* set maximal menu item width */

  /* --- determine controls width --- */
  #ifdef LINUX                  /* if Linux Athena Widgets */
  XtVaGetValues(psh->form, XtNdefaultDistance, &dd, NULL);
  #else                         /* get the default distance, */
  dd = 4;                       /* if Sun or HP Athena Widgets */
  #endif                        /* set the default distance */
  psi = psh->items; max = wd = 0;  /* traverse property sheet items */
  for (n = psh->itemcnt; --n >= 0; psi++) {
    if (psi->type == PS_LINE) { /* skip labels, */
      if (wd > max) max = wd;   /* if last control line was wider, */
      wd = 0; continue;         /* note new maximal width, reset */
    }                           /* the width and continue */
    XtVaGetValues(psi->widget, XtNwidth, &tmp, NULL);
    wd += (wd > 0) ? tmp +dd +2 : tmp;
  }                             /* determine the control width sum */
  total += (wd > max) ? wd:max; /* determine total width of box and */
  wd = (psh->ca) ? (total -2) /2 : total +dd;       /* button width */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNwidth,    wd);            n++;
  XtSetArg(args[n], XtNfromVert, psh->currline); n++;
  XtSetValues(psh->ok, args, n);/* set width of ok button */
  if (psh->ca) {                /* if there is a cancel button */
    XtSetArg(args[0], XtNwidth, total -2 -wd);
    XtSetValues(psh->ca, args, n);
  }                             /* set width of cancel button */

  /* --- set width of line controls --- */
  psi = psh->items;             /* traverse property sheet items */
  for (n = psh->itemcnt; --n >= 0; psi++) {
    if ((psi->type != PS_LINE)  /* skip non-labels */
    ||   psi->widget) continue; /* and non-empty labels */
    XtVaSetValues(psi[1].widget, XtNwidth, total +dd, NULL);
  }                             /* set maximal label width */

  /* --- set callback functions --- */
  XtRemoveAllCallbacks(psh->ok, XtNcallback);
  XtAddCallback(psh->ok, XtNcallback, cb_ok, (XtPointer)callback);
  if (psh->ca) {                /* if there is a cancel button */
    XtRemoveAllCallbacks(psh->ca, XtNcallback);
    XtAddCallback(psh->ca, XtNcallback, cb_cancel, (XtPointer)callback);
  }                             /* set callback function for it, too */

  /* --- install actions and accelerators --- */
  XtAppAddActions(XtWidgetToApplicationContext(psh->parent),
    actions, XtNumber(actions));
  XtInstallAllAccelerators(psh->form, psh->form);

  /* --- show dialog box --- */
  XtVaGetValues(psh->parent, XtNx, &x, XtNy, &y, NULL);
  XtVaSetValues(psh->shell,  XtNx,  x, XtNy,  y, NULL);
  XtPopup(psh->shell, XtGrabExclusive);
  for (psi = psh->items, n = 0; n < psh->itemcnt; n++, psi++) {
    if (!psi->widget) continue; /* traverse property sheet items */ 
    if (XtClass(psi->widget) == asciiTextWidgetClass) {
      XtSetKeyboardFocus(psh->shell, psi->widget); break; }
  }                             /* set keyboard focus */

  XtVaGetValues(psh->form, XtNdefaultDistance, &dd, NULL);

  /* --- set position and inhibit resize --- */
  XtVaGetValues(psh->shell, XtNwidth,   &wd, XtNheight,   &ht, NULL);
  XtVaSetValues(psh->shell, XtNminWidth, wd, XtNminHeight, ht,
                            XtNmaxWidth, wd, XtNmaxHeight, ht, NULL);

  /* --- override window manager delete --- */
  XtOverrideTranslations(psh->shell,
    XtParseTranslationTable("<Message>WM_PROTOCOLS: db_close()"));
  wm_delete = XInternAtom(XtDisplay(psh->shell),
    "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(psh->shell), XtWindow(psh->shell),
    &wm_delete, 1);             /* override window manager delete */
  return 0;                     /* return `ok' */
}  /* ps_handle() */
